from sklearn.model_selection import cross_val_score
from sklearn import metrics, linear_model
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
from sklearn.pipeline import Pipeline
from sklearn.utils import resample
import numpy as np
import pandas as pd
import math
from tqdm import tqdm

class FeatureSelector(object):
    def __init__(self, data, target, 
                 C=None, C_space=np.linspace(0.0001, 1, 20), C_finder_iter=100, C_tol = 0.005, 
                 cut_off_frac_model=0, cut_off_frac_estimation=True, cut_off_estim_params=None,
                 pipeline_steps=None, scoring='roc_auc'):
        """
        Feature selection using regularized logistic regression with automatic C parameter tuning.
        
        Parameters
        ----------
        data : array-like or pd.DataFrame
            Feature matrix (n_samples, n_features)
        target : array-like
            Binary target vector (n_samples,)
        C : float, optional
            Regularization strength (inverse of regularization). 
            If None, will be automatically determined from C_space.
        C_space : array-like, optional
            Search space for C values (default: 20 values from 0.0001 to 1)
        C_finder_iter : int, optional
            Number of bootstrap iterations for C optimization (default: 100)
        C_tol : float, optional
            Tolerance for derivative to consider the optimization plateau (default: 0.005)
        cut_off_frac_model : float, optional
            Fraction cutoff for model coefficients (default: 0)
        cut_off_frac_estimation : bool, optional
            Whether to estimate optimal cutoff fraction (default: False)
        cut_off_estim_params : dict, optional
            Parameters for cutoff fraction estimation. If None, uses:
            {'inner_loop': 10, 'max_iter': 10, 'cut_off_feature_value': 0.1, 'max_feature': None, optimal_method': 'first'}
            inner_loop:
                The number of simulations of the inner loop when evaluating cut_off_level. The number of repetitions of train_test_splits to evaluate the quality of a model with a given set of features.
            max_iter:
                Number of simulations when adding each feature during cut-off search. The total number of simulations is max_iter*inner_loop.
            cut_off_feature_value:
                The minimum proportion of models that a feature must be included in for it to be considered in further simulations
            max_feature : int, optional 
                Can be restrict the feature space to n-top features. If "None", then the search is carried out over the entire feature space. 
            optimal_method': str, ['first', 'median']
                How optimal cut-off level will be get. Default = 'first'.
        pipeline_steps : list, optional
            Custom pipeline steps for the classifier. If None, uses:
            [StandardScaler(), LogisticRegression(penalty='l1', solver='liblinear')]
        scoring : str, optional
            Scoring metric for model evaluation. Must be from sklearn.metrics.SCORERS.keys()
            (default: 'roc_auc')
            
        Notes
        -----
        - Uses L1 regularization (Lasso) for automatic feature selection
        - When C=None, performs bootstrap-based optimization to find optimal regularization strength
        - Final model selects features with non-zero coefficients
        """
        
        self.best_features = dict()
        self.score_list = list()
        self.best_feature_progress = list()
        self.best_features_log = dict()
        self.old_top = set()
        self.top_update = list()
        self.top_weights = list()
        self.top_fearures_progress = list()
        self.n_model = 0
        self.C_finder_iter = C_finder_iter
        self.C_space = C_space
        self.data = data
        self.target = target
        self.plot_params = dict()
        self.plot_params['cutoff_searching'] = {'cut_of_list': None, 
                                                'roc_train_list': None, 
                                                'roc_test_list': None, 
                                                'arg_max_list': None}
        self.cut_off_frac_model = cut_off_frac_model
        self.scoring = scoring
        self.C_space_iter = None

        # Default parameters for cut_off_estim_params
        default_cut_off_params = {
            'inner_loop': 10,
            'max_iter': 10,
            'cut_off_feature_value': 0.1,
            'max_feature': None, # Feature space can be restrict to n-top features. If "None", then the search is carried out over the entire feature space. 
            'optimal_method': 'first' 
        }
        if cut_off_estim_params is not None:
            default_cut_off_params.update(cut_off_estim_params)

        # Defining pipeline steps
        if not C:
            print('The regularization coefficient was not specified, the search for the optimal C was started')
            self.C = self.get_optimal_C(tol=C_tol)
            print('Optimal regularization coefficient (С) = ', round(self.C, 3))
        else:
            self.C = C
        if pipeline_steps:
            self.pipeline = Pipeline(pipeline_steps)
        else:
            steps = [
                ('scaler', StandardScaler()),
                ('log_reg', linear_model.LogisticRegression(penalty='l1', C=self.C, solver='liblinear', random_state=42))
            ]
            # Creating a pipeline
            self.pipeline = Pipeline(steps)
        if cut_off_frac_estimation:
            print()
            print('Prefit model for cutoff weight level estimation')
            self.fit()
            print('Prefit done')
            self.cut_off_frac_model = self._get_optimal_cut_off_level(**cut_off_estim_params)

    def get_optimal_C(self, tol=0.005):
        """
        The function searches for the optimal regularization parameter
        """
        X = self.data
        y = self.target
        
        result_df_train = pd.DataFrame()

        for a in tqdm(range(0, self.C_finder_iter), desc='processing: '):
            for _ in range(self.C_finder_iter):
                X_train, _, y_train, _ = train_test_split_res(X, y)
                if count_min_class(y_train) > 3:
                    break
            roc_auc_train = list()
            for i in self.C_space:
                steps = [
                    ('scaler', StandardScaler()),
                    ('log_reg', linear_model.LogisticRegression(penalty='l1', C=i, solver='liblinear', random_state=42))
                ]
                # Creating a pipeline
                pipeline = Pipeline(steps)
                pipeline.fit(X_train, y_train)        
                roc_auc_train.append(np.mean(cross_val_score(pipeline, X_train, y_train, cv=3, scoring=self.scoring)))
            result_df_train[a] = roc_auc_train

        _y = result_df_train.apply(np.mean, axis=1)
        _x = range(0, len(_y))

        der_y = np.diff(_y)
        max_der_y = np.where(der_y == max(der_y))[0][0]
        arg_plateau = max_der_y + np.where(der_y[max_der_y:] < tol)[0][0]

        self.C_space_mean = _y
        self.C_space_iter = _x
        self.C_arg_plateau = arg_plateau

        return self.C_space[arg_plateau]

    def fit(self,
            max_iter: int = 3000,
            cut_off_score: float = 0.6,
            log: bool = True,
            feature_resample: int = 0):
        """
        Train the feature selection model and identify significant features.
        
        Parameters
        ----------
        max_iter : int, optional (default=3000)
            Maximum number of iterations for the optimization solver.
            Must be positive. Consider increasing for complex datasets.
            
        cut_off_score : float, optional (default=0.6)
            Minimum importance score (scorer_metrics [ROC-AUC] * proportion-of-models-that-include-the-feature) threshold for feature selection (range: 0-1).
            Features with scores below this threshold will be discarded.
            
        log : bool, optional (default=True)
            If True, enables verbose output during training.
            Recommended for monitoring convergence.
            
        feature_resample : int, optional (default=0)
            Resampling of features. If 0, the full feature space is considered at each train_test_split, 
            otherwise the feature space is also sampled in batches of size "feature_resample"

        Returns
        -------
        self : FeatureSelector
            The fitted estimator instance enabling method chaining.

        Examples
        --------
        >>> selector = FeatureSelector(X, y)
        >>> selector.fit(max_iter=5000)
        >>> selected_features = selector.get_support()
        """
        roc_auc_scorer = metrics.make_scorer(metrics.roc_auc_score)
        self.best_features = dict()
        self.best_feature_progress = list()
        self.best_features_log = dict()
        self.top_update = list()
        self.top_weights = list()
        self.top_fearures_progress = list()
        self.old_top = set()
        self.n_model = 0
        self.best_features_counter = dict()

        for itr in tqdm(range(1, max_iter+1), desc='fit model: '):
            # model
            for _ in range(max_iter):
                X_train, X_test, y_train, y_test = train_test_split_res(self.data, self.target, feature_resample=feature_resample)
                if count_min_class(y_train) > 3:
                    if len(set(y_test)) > 1:
                        break
            if feature_resample:                
                for i in X_train.columns:
                    if i in self.best_features_counter:
                        self.best_features_counter[i] += 1
                    else:
                        self.best_features_counter[i] = 1
            self.pipeline.fit(X_train, y_train)
            mask = self.pipeline.named_steps['log_reg'].coef_[0] != 0
            score = roc_auc_scorer(self.pipeline, X_test, y_test)
            self.score_list.append(roc_auc_scorer(self.pipeline, X_test, y_test))
            if score < cut_off_score or np.isnan(score):
                continue
            else:
                self.n_model += 1
#            for i in self.data.columns[mask]:
            for i in X_train.columns[mask]:
                if i in self.best_features:
                    self.best_features[i]+=1*score
                    self.best_features_log[i]+=1
                else:
                    self.best_features[i]=1*score
                    self.best_features_log[i]=1

            # loging
            if log:
                if feature_resample:
                    top_feat = set([k for k, v in self.best_features.items() if v/self.best_features_counter[k] > self.cut_off_frac_model])
                else:
                    top_feat = set([k for k, v in self.best_features.items() if v/self.n_model > self.cut_off_frac_model])
                ft_prog = len(top_feat)
                if top_feat - self.old_top:
                    self.top_update.append(1)
                else:
                    self.top_update.append(0)
                self.old_top = top_feat
                self.best_feature_progress.append(len(self.best_features))
                self.top_fearures_progress.append(ft_prog)
                self.top_weights.append(sum([v for k, v in self.best_features.items() if k in top_feat]))

        for  k, v in self.best_features.items():
            if feature_resample:
                self.best_features[k] = v/self.best_features_counter[k]
            else:
                self.best_features[k] = v/self.n_model
        
        self.all_features = dict(sorted(self.best_features.items(), key=lambda item: item[1], reverse=True))
        self.best_features = {k:v for k, v in self.all_features.items() if v > self.cut_off_frac_model}

    def _get_optimal_cut_off_level(self, X=None, y=None, best_features=None, self_data=True, inner_loop=20, max_iter=10, cut_off_feature_value=0.1, max_feature=None, optimal_method='first'):
            
            _cut_of_list = list()
            _roc_train_list = list()
            _roc_test_list = list()
            _arg_max_list = list()
            if self_data:                    
                X = self.data
                y = self.target
            else:
                self.best_features = best_features
            roc_auc_scorer = metrics.make_scorer(metrics.roc_auc_score)
            
            print('Serching cutoff level for feature weights...', end=' ')
            for i in range(max_iter):
                print(i, end=' ')
                sorted_feature = {k: v for k, v in sorted(self.best_features.items(), 
                                                        key=lambda item: item[1], reverse=True)}

                rocauc_mean_train = list()
                rocauc_mean_test = list()
                _feature = list()
                for i in tqdm(sorted_feature.keys(), desc='feature space analysis: '):
                    if sorted_feature[i] < cut_off_feature_value:
                        break
                    roc_auc_train = list()
                    roc_auc_test = list()
                    _feature.append(i)
                    if max_feature:
                        if len(_feature) > max_feature:
                            break
                    tmp_X = X.loc[:,_feature]
                    
                    for _ in range(inner_loop):
                        for _ in range(inner_loop):
                            X_train, X_test, y_train, y_test = train_test_split_res(tmp_X, y)
                            if count_min_class(y_train) > 3: # so that the number of observations in the minimum class is greater than the number of splits in the CV
                                if len(set(y_test)) > 1:
                                    break
                        steps = [
                            ('scaler', StandardScaler()),
                            ('log_reg', linear_model.LogisticRegression(penalty='l2', C=1, solver='liblinear', random_state=42))
                        ]
                        # Creating a pipeline
                        pipeline = Pipeline(steps)
                        pipeline.fit(X_train, y_train)        
                        roc_auc_train.append(np.mean(cross_val_score(pipeline, X_train, y_train, cv=3, scoring=self.scoring)))
                        roc_auc_test.append(roc_auc_scorer(pipeline, X_test, y_test))
                    rocauc_mean_train.append(np.mean(roc_auc_train))
                    rocauc_mean_test.append(np.mean(roc_auc_test))
                
                rc = np.array(rocauc_mean_test)
                arg_max = np.where(rc == max(rc))[0][0]
                last_feature = list(sorted_feature.keys())[arg_max]
                cut_off_level = math.floor((sorted_feature[last_feature])*100)/100
                _cut_of_list.append(cut_off_level)
                _roc_train_list.append(rocauc_mean_train)
                _roc_test_list.append(rocauc_mean_test)
                _arg_max_list.append(arg_max)
                
            self.plot_params['cutoff_searching'] = {'cut_of_list': _cut_of_list, 
                                                    'roc_train_list': _roc_train_list, 
                                                    'roc_test_list': _roc_test_list, 
                                                    'arg_max_list': _arg_max_list}
            if optimal_method=='first':
                optimal_cut_off = sorted(_cut_of_list, reverse=True)[0]
            if optimal_method=='median':
                optimal_cut_off = np.median(_cut_of_list)
            print('optimal cut of weight level = ', optimal_cut_off)
            return optimal_cut_off
        
class fsplot(object):
    def __init__(self, fs_model, color = [(18/255, 163/255, 173/255, 1), #"green"
                                          (200/255, 13/255, 63/255, 1),  #"red"
                                          (248/255, 155/255, 54/255, 1), #the third color is orange
                                          (242/255, 104/255, 73/255, 1), #fourth color
                                          (65/255, 100/255, 175/255, 1)] #fifth color
                ) -> None:
        
        self.one_color = color[4]
        self.two_color = color[1]
        self.third_color = color[2]
        self.color = color
        self.fs_model = fs_model

    def ROCdistr(self, ax=None, fontsize=20, path=None):
        if ax:
            ax1 = ax
        else:
            _, ax1 = plt.subplots(figsize=(12, 10))
        
        ax1.hist(self.fs_model.score_list, bins=15, color=self.two_color, alpha=1., edgecolor='black')
        ax1.set_xlabel('ROC-AUC', fontsize=fontsize)
        ax1.set_ylabel('Number of models', fontsize=fontsize)
        ax1.tick_params(labelsize=fontsize)
        ax1.yaxis.labelpad = fontsize  # Y-axis offset
        ax1.xaxis.labelpad = fontsize  # X-axis offset
        # Remove the top and right borders of the axes
        ax1.spines['top'].set_visible(False)
        ax1.spines['right'].set_visible(False)
        
        # Save the histogram to a file
        if path:
            plt.savefig(path, dpi=300, bbox_inches='tight')
    
    def feature_selection(self, ax=None, fontsize=20, path=None):
        if ax:
            ax1 = ax
        else:
            _, ax1 = plt.subplots(figsize=(12, 10))
        top_ind = np.where(np.diff(self.fs_model.top_update))[0]
        for index in top_ind:
            ax1.vlines(index, 0, 0.5, color=self.two_color, linestyles='dashed', linewidth=4)
        ax1.plot(np.array(self.fs_model.top_fearures_progress[2:]), color=self.third_color, linewidth=4)
        ax1.set_xlabel('Iteration number', fontsize=fontsize)
        ax1.set_ylabel('Number of features included \n in more than {}% models'.format(round(self.fs_model.cut_off_frac_model*100)), fontsize=fontsize)
        ax1.tick_params(labelsize=fontsize)
        ax1.yaxis.labelpad = fontsize  # Y-axis offset
        ax1.xaxis.labelpad = fontsize  # X-axis offset
        ax1.spines['top'].set_visible(False)
        ax1.spines['right'].set_visible(False)

        if path:
            plt.savefig(path, dpi=300, bbox_inches='tight')
        
    def feature_weights(self, ax=None, fontsize=20, path=None):
        if ax:
            ax1 = ax
        else:
            _, ax1 = plt.subplots(figsize=(12, 10))
        value = [v for k, v in self.fs_model.all_features.items()]
        ax1.hist(value, bins=19, color=self.third_color, alpha=1., edgecolor='black')
        ax1.vlines(self.fs_model.cut_off_frac_model, 0, 20, color=self.two_color, linewidth=4)
        ax1.text(self.fs_model.cut_off_frac_model/max(value)+0.02, 0.3, 'cutoff\nlevel = {}'.format(np.round(self.fs_model.cut_off_frac_model, 2)), transform=ax1.transAxes, color=self.two_color, fontsize=fontsize, ha='left', va='center')
        ax1.set_xlabel('Estimated weight', fontsize=fontsize)
        ax1.set_ylabel('Number of features', fontsize=fontsize)
        ax1.tick_params(labelsize=fontsize)
        ax1.yaxis.labelpad = fontsize  # Y-axis offset
        ax1.xaxis.labelpad = fontsize  # X-axis offset
        ax1.spines['top'].set_visible(False)
        ax1.spines['right'].set_visible(False)

        if path:
            plt.savefig(path, dpi=300, bbox_inches='tight')
        
    def feature_selection_exted(self, ax=None, fontsize=20, path=None):
        # Creating a graph
        if ax:
            ax1 = ax
        else:
            fig, ax1 = plt.subplots(figsize=(12, 10))
        # Plotting the first curve
        ax1.plot(np.array(self.fs_model.best_feature_progress), color=self.one_color, linewidth=4)
        ax1.set_xlabel('Iteration number', fontsize=fontsize)
        ax1.set_ylabel('Number of features \n included in at least one model', color=self.one_color, fontsize=fontsize)
        ax1.tick_params(axis='y', colors=self.one_color)
        ax1.tick_params(labelsize=fontsize)
        ax1.yaxis.labelpad = fontsize  # Y-axis offset
        ax1.xaxis.labelpad = fontsize  # X-axis offset
        ax1.spines['top'].set_visible(False)
        ax1.spines['right'].set_visible(False)

        # Create a second y-axis
        ax2 = ax1.twinx()
        ax2.plot(np.array(self.fs_model.top_weights), color=self.third_color, linewidth=4)
        ax2.set_ylabel('Total weight of features \n included in more than {} of models'.format(round(self.fs_model.cut_off_frac_model)), color='black', fontsize=fontsize)
        ax1.tick_params(labelsize=fontsize)
        ax1.yaxis.labelpad = fontsize  # Y-axis offset
        ax1.xaxis.labelpad = fontsize  # X-axis offset
        ax1.spines['top'].set_visible(False)
        ax1.spines['right'].set_visible(False)

        top_ind = np.where(np.diff(self.fs_model.top_fearures_progress))[0]
        for index in top_ind:
            ax1.vlines(index, 0, 1.7, color=self.two_color, linestyles='dashed', linewidth=4)
        if path:
            plt.savefig(path, dpi=300, bbox_inches='tight')
    
    def C_searching(self, ax=None, fontsize=20, path=None):
        # Creating a graph
        if ax:
            ax1 = ax
        else:
            _, ax1 = plt.subplots(figsize=(12, 10))
        ax1.plot(self.fs_model.C_space_iter, self.fs_model.C_space_mean, color=self.one_color, label='train', linewidth=4)
        ax1.vlines(self.fs_model.C_arg_plateau, min(self.fs_model.C_space_mean), max(self.fs_model.C_space_mean), color=self.two_color, linewidth=4)
        ax1.text(self.fs_model.C_arg_plateau/max(self.fs_model.C_space_iter)+0.03, 0.2, 'optimal\nС = {}'.format(np.round(self.fs_model.C,2)), transform=ax1.transAxes, color=self.two_color, fontsize=fontsize, ha='left', va='center')
        ax1.set_xlabel('Inverse regularization coefficient (С)', fontsize=fontsize)
        ax1.set_ylabel('ROC-AUC train CV', fontsize=fontsize)
        ax1.set_xticks(ticks=self.fs_model.C_space_iter[::4])
        ax1.set_xticklabels(labels=np.round(self.fs_model.C_space[::4],2))
        ax1.tick_params(labelsize=fontsize)
        ax1.yaxis.labelpad = fontsize  # Y-axis offset
        ax1.xaxis.labelpad = fontsize  # X-axis offset
        ax1.spines['top'].set_visible(False)
        ax1.spines['right'].set_visible(False)
        

        if path:
            plt.savefig(path, dpi=300, bbox_inches='tight')
    
    def weights_cut_of_level_searching(self, ax=None, fontsize=20, path=None):
        if ax:
            ax1 = ax
        else:
            _, ax1 = plt.subplots(figsize=(12, 10))

        train_series = np.array(self.fs_model.plot_params['cutoff_searching']['roc_train_list'])
        test_series = np.array(self.fs_model.plot_params['cutoff_searching']['roc_test_list'])
        arg_max_list = np.array(self.fs_model.plot_params['cutoff_searching']['arg_max_list'])
        color_line = [self.one_color, self.two_color] #green, red
        color_hist = self.third_color
        legend = ['train', 'test']

        args = self.fs_model.plot_params['cutoff_searching']['arg_max_list']
        cutoff = self.fs_model.plot_params['cutoff_searching']['cut_of_list']
        dct_labels = dict(zip(args, cutoff))

        s_key = sorted(dct_labels.keys())
        s_y = np.linspace(0.4,0.05,len(dct_labels))
        d_y = dict(zip(s_key, s_y))
        y_text = [d_y[i] for i in dct_labels.keys()]
        #Plotting:
        ax1.hist(arg_max_list, alpha=.7, color=color_hist, density=True)
        ax1.vlines(dct_labels.keys(), 0, y_text, colors=self.two_color)
        # Adding labels above lines
        i=0
        for x, label in dct_labels.items():
            ax1.text(x, y_text[i], label, ha='center', va='bottom', fontsize=(fontsize-9))
            i+=1
        ax2 = ax1.twinx()

        for i, x in enumerate([train_series, test_series]):
            smooth_path    = x.mean(axis=0)
            path_deviation = 2.3 * x.std(axis=0)
            under_line     = (smooth_path-path_deviation)
            over_line      = (smooth_path+path_deviation)
            ax2.plot(smooth_path, linewidth=4, color=color_line[i], label=legend[i]) #mean curve.
            ax2.fill_between(range(smooth_path.shape[0]), under_line, over_line, color=color_line[i], alpha=.1) #std curves.

        ax2.set_ylabel('ROC-AUC', color=self.two_color, fontsize=fontsize)
        ax2.tick_params(axis='y', colors=self.two_color)
        ax1.set_ylabel('Share of simulations', color='black', fontsize=fontsize)
        ax1.set_xlabel('Number of features included in the model', color='black', fontsize=fontsize)
        ax1.tick_params(labelsize=fontsize)
        ax2.tick_params(labelsize=fontsize)
        ax1.yaxis.labelpad = fontsize  
        ax1.xaxis.labelpad = fontsize  
        ax1.spines['top'].set_visible(False)
        ax1.spines['right'].set_visible(False)
        ax1.yaxis.labelpad = fontsize  
        ax2.yaxis.labelpad = fontsize  
        ax1.xaxis.labelpad = fontsize  
        ax2.legend(fontsize=fontsize, loc='lower right')
        
        if path:
            plt.savefig(path, dpi=300, bbox_inches='tight')
        pass
        
    def plot_all(self, fontsize=20, labels = ['a.', 'b.', 'c.', 'd.', 'e.', 'f.'], left=0.1, right=0.9, top=0.9, bottom=0.1, hspace=0.3, wspace=0.4):
        _, axs = plt.subplots(3, 2, figsize=(24, 30))
        if self.fs_model.C_space_iter:
            self.C_searching(ax=axs[0, 0], fontsize=fontsize)
        else:
            axs[0, 0].text(0.5, 0.5, f"No search for C was performed.\nC = {self.fs_model.C}", 
                           ha='center', va='center', fontsize=fontsize)
        self.ROCdistr(ax=axs[0, 1], fontsize=fontsize)
        if self.fs_model.plot_params['cutoff_searching']['roc_train_list']:
            self.weights_cut_of_level_searching(ax=axs[1, 0], fontsize=fontsize)
        else:
            axs[1, 0].text(0.5, 0.5, f"No cut_of_level search was performed.\ncut_of_level = {self.fs_model.cut_off_frac_model}", 
                           ha='center', va='center', fontsize=fontsize)
        self.feature_weights(ax=axs[1, 1], fontsize=fontsize)
        self.feature_selection(ax=axs[2, 0], fontsize=fontsize)
        self.feature_selection_exted(ax=axs[2, 1], fontsize=fontsize)
        
        # Add letter numbering outside of graphs
        for i, ax in enumerate(axs.flat):
            # Define positions for annotations
            if i % 2 == 0:
                # Left graphs
                x_pos = -.1
            else:
                # Right graphs
                x_pos = -.1
            y_pos = 1.1
            
            ax.annotate(labels[i], xy=(x_pos, y_pos), xycoords='axes fraction', fontsize=fontsize+5,
                        ha='center', va='center', bbox=dict(facecolor='white', edgecolor='none', pad=1))
        plt.tight_layout()
        plt.subplots_adjust(left=left, right=right, top=top, bottom=bottom, hspace=hspace, wspace=wspace)

def train_test_split_res(X, y, feature_resample=0):
    """
    train test split with resampling
    """
    N = len(y)
    # resample each class separately to keep their proportions
    train_indx = list(resample(np.where(y)[0])) + list(resample(np.where(np.array(y) == 0)[0]))
    test_indx = list(set(range(N)).difference(train_indx))
    X_train = X.iloc[train_indx,:]
    X_test = X.iloc[test_indx,:]
    y_train = [y[i] for i in train_indx]                    
    y_test = [y[i] for i in test_indx]
    if feature_resample:
        random_columns = np.random.choice(X_train.columns, feature_resample, replace=False)
        X_train = X_train[random_columns]
        X_test = X_test[random_columns]
    return X_train, X_test, y_train, y_test

def get_feature_space(model_list, cut_off_level=0.25):
    """
    the function returns a set of FeatureSelector models that satisfy the condition
    """
    fs = list()
    for model in model_list:
        fs = fs + [k for k, v in model.all_features.items() if v > cut_off_level]
    return set(fs)

def count_min_class(vector):
    # Convert the list to a NumPy array if it is not already a NumPy array
    vector = np.array(vector)
    # Count the number of 0 and 1
    counts = np.bincount(vector)
    # Return the number of the smallest class
    return np.min(counts)


import networkx as nx  # version == 2.3
import pandas as pd
import requests  # HTTP Client for Python
import json  # Standard JSON library
from py2cytoscape import util as cy
import matplotlib.pyplot as plt
from scipy.ndimage import gaussian_filter1d
import numpy as np
from colour import Color
from itertools import combinations
from importlib.resources import files
#   load interactome from String and convert to nx Graph


#DATAPATH = files("genelens").joinpath("data/miRNET/")

class MainNet:

    def __init__(self, interactome=None):
        if not isinstance(interactome, pd.DataFrame):
            string = pd.read_csv(files("genelens").joinpath("data/miRNET/baseData/String_interactome.csv"))
            self.G = nx.from_pandas_edgelist(string, 'Source', 'Target')
        else:
            interactome = pd.DataFrame(interactome)
            assert interactome.shape[1] == 2, 'It takes two columns: "Source" and "Target"'
            assert sum(interactome.columns == ['Source', 'Target']) == 2, 'Columns names are not "Source" and "Target"'
            self.G = nx.from_pandas_edgelist(interactome, 'Source', 'Target')
            print('interactome contain ', interactome.shape[0], ' rows')

    def get_LCCnd_centrality(self):
        """
        return: sorted dict of node centrality
        """
        centrality_node = nx.betweenness_centrality(self.get_LCC())
        degree_centrality = nx.degree_centrality(self.get_LCC())
        for k, v in centrality_node.items():
            centrality_node[k] = centrality_node[k] + degree_centrality[k]
        centrality_node = {k: v for k, v in sorted(centrality_node.items(), key=lambda item: item[1], reverse=True)}

        for nods, dct in self.LCC.nodes(data=True):
            dct['BtweenCentrl'] = centrality_node[nods]

        return centrality_node

    def get_LCC(self):
        """
        return: the Largest Connected Component, as NetrworkX object
        """
        CC_G = [self.G.subgraph(c).copy() for c in nx.connected_components(self.G)]
        self.LCC = max(CC_G, key=len)
        
        #old version NetworkX
        #CC_G = sorted(nx.connected_component_subgraphs(self.G), key=len, reverse=True)
        #self.LCC = CC_G[0]


        # print(len(CC_G), 'total connected components')
        # print(len(CC_G[0].nodes), 'LCC cardinality')
        
        print('LCC was extracted')
        print(f"Total connected components={len(CC_G)}, LCC cardinality={len(self.LCC)}")

    def select_nodes(self, gene_set):
        """
        param gene_set: tissue or another gene set
        """

        self.G = self.G.subgraph(gene_set)
        if self.LCC:
            self.LCC = self.LCC.subgraph(gene_set)
    
    def minimum_connected_subgraph(self, required_nodes):
        """
        Finds the minimal connected subgraph containing the specified nodes.
        
        This implementation computes the smallest subgraph (by edge count) that:
        1. Contains all nodes from the input set
        2. Maintains connectivity between all included nodes
        
        Parameters
        ----------
        required_nodes : Union[List, Set]
        Prespecified nodes that must be included in the subgraph.
        Can be provided as either a list or set of node identifiers.
        """
        req_top_gene_dict = required_nodes.copy()
        for gene in required_nodes.keys():
            if gene not in self.LCC.nodes():
                req_top_gene_dict.pop(gene)
                print(f'{gene} absent from LCC, excluded from further analysis')

        # Create an auxiliary graph between the given vertices
        auxiliary_graph = nx.Graph()
        for u, v in combinations(req_top_gene_dict.keys(), 2):
            # Find the length of the shortest path and its edges
            path_length = nx.shortest_path_length(self.LCC, source=u, target=v, weight='weight')
            auxiliary_graph.add_edge(u, v, weight=path_length)
        
        # Construct a minimum spanning tree of the auxiliary graph
        mst = nx.minimum_spanning_tree(auxiliary_graph, weight='weight')
        
        # Expand MST edges in the original graph to a complete subgraph
        subgraph_edges = set()
        for u, v in mst.edges:
            path = nx.shortest_path(self.LCC, source=u, target=v, weight='weight')
            subgraph_edges.update(zip(path[:-1], path[1:]))

        self.mst_subgraph = self.LCC.edge_subgraph(subgraph_edges).copy()
        
        print()
        print('mst-graph was extracted')
        print(f"Initial core feature={len(req_top_gene_dict.keys())}, mst-graph cardinality={len(self.mst_subgraph)}")


def tissue_selector(ans=None, tissue_id=None):
    
    if ans is None:
        ans = int(str(input('"Human Protein Atlas"(0) or "GTEx"(1) ? ')))
    if ans == 1:
        dt = pd.read_csv(
            files("genelens").joinpath("data/miRNET/addData/GTEx_Analysis_2017-06-05_v8_RNASeQCv1.1.9_gene_median_tpm.gct"),
            sep='\t')  # loading med(TPM) from GTEx

        print('Gene universe is...')

        labels = sorted(dt.columns)
        index = [i for i in range(0, len(dt.columns))]

        if tissue_id is None:
            for i in index:
                print(index[i], '-----', labels[i]) 
            print("give me int, or input 'all'", sep='\n')
            tissue_id = str(input('Your choice: '))
        else:
            print(labels[int(tissue_id)], 'was used', sep=' ')

        if tissue_id != 'all':
            tissue = labels[int(tissue_id)]
            tissue_genes = set(dt['Description'][dt[tissue] > 0])
            print('your tissue is ', tissue, ' number of genes: ', len(tissue_genes))

            return tissue_genes
        else:
            return 'all'

    elif ans == 0:
        dt = pd.read_csv(
            files("genelens").joinpath("data/miRNET/addData/Hum_Atas_normal_tissue.tsv"),
            sep='\t')
        print('Gene universe is...')

        labels = sorted(list(map(str, set(dt['Tissue']))))
        index = [i for i in range(0, len(labels))]
        if tissue_id is None:
            for i in index:
                print(index[i], '-----', labels[i])
            print("give me int, or input 'all'", sep='\n')
            tissue_id = str(input('Your choice: '))
        else:
            print(labels[int(tissue_id)], 'was used', sep=' ')

        if tissue_id != 'all':
            tissue = labels[int(tissue_id)]
            tissue_genes = set(dt[(dt['Tissue'] == tissue) & (dt['Level'] != 'Not detected')]['Gene name'])
            print('your tissue is ', tissue, ' number of genes: ', len(tissue_genes))
            return tissue_genes
        else:
            return 'all'
    else:
        return tissue_selector()


class KeyNodesExtractor:

    def __init__(self, MirNet):
        self.key_nodes = dict()
        self.net = MirNet.get_LCC()
        self.node_centrality = MirNet.get_LCCnd_centrality()
        self.graph_features = {'card_LCC': [len(self.net.nodes())],
                               'n_CC': [len(list(nx.connected_component_subgraphs(self.net)))],
                               'transitivity': [nx.transitivity(self.net)],
                               'sh_path': [nx.average_shortest_path_length(self.net) / len(self.net.nodes())]}

    @staticmethod
    def inflection_finder(card_LCC, n_CC, sigma):
        """
        :param sigma: smoothing
        :param card_LCC: cardinality of the LCC
        :param n_CC: number of connected components in the Network
        :return: the index of the last key node, after the removal of which the network stops rapidly falling apart
        """

        y = gaussian_filter1d(card_LCC, sigma=sigma)
        dy = np.diff(y)  # first derivative
        idx_max_dy = np.argmax(dy)
        if card_LCC[idx_max_dy] > n_CC[idx_max_dy]:
            return KeyNodesExtractor.inflection_finder(card_LCC, n_CC, sigma + 0.2)
        else:
            return idx_max_dy

    def extraction(self):

        for k, v in self.node_centrality.items():

            if v == 0:
                break
            self.net.remove_node(k)
            if len(self.net.nodes()) == 0:
                break
            LCC_curent = sorted(nx.connected_component_subgraphs(self.net), key=len, reverse=True)[0]   # get LCC
            self.graph_features['card_LCC'].append(len(LCC_curent.nodes()))
            self.graph_features['n_CC'].append(len(list(nx.connected_component_subgraphs(self.net))))
            self.graph_features['transitivity'].append(nx.transitivity(LCC_curent))
            self.graph_features['sh_path'].append(nx.average_shortest_path_length(LCC_curent) / len(LCC_curent.nodes()))

        # find inflection point of function

        idx_max_dy = KeyNodesExtractor.inflection_finder(card_LCC=self.graph_features['card_LCC'],
                                                         n_CC=self.graph_features['n_CC'],
                                                         sigma=0.0001)

        self.graph_features['cutoff_point'] = idx_max_dy

        for i in range(0, idx_max_dy + 1):
            nods = list(self.node_centrality.keys())[i]
            self.key_nodes[nods] = self.node_centrality[nods]

        return self.key_nodes.keys()


class Targets:
    """
    The class extracts and stores the targets of one microRNA and its name
    """

    def __init__(self, path_to_miRTarBase):
        self.miR_dict = {}
        with open(path_to_miRTarBase) as interact:  # import targets from miRTarBase
            for line in interact:
                (key, val) = line.strip().split(';')
                if key in self.miR_dict:
                    self.miR_dict[key] += [val]
                else:
                    self.miR_dict[key] = [val]

    def get_targets(self, miR_name):
        """
        miRTarBase contains different forms of miRNA, for example, miR-21- can correspond to
        miR-21-3p and miR-21-5p. The function concatenates targets of all forms of miRNA and removes duplicates,
        which arise because the same target can correspond to several lines due to
        different methods of its confirmation.
        """

        miR_names = self.miR_dict.keys()
        miR_dict = self.miR_dict

        res = set()
        mir_name_app = []

        for name in miR_names:
            if miR_name in name:
                res.update(miR_dict[name])
                mir_name_app.append(name)

        if not mir_name_app:
            print('miRNA', '"{}"'.format(miR_name), 'not found, use another name')
            return 1

        print('I found a miRNA with name:', *mir_name_app)
        print('and ', len(res), 'unique targets')

        return res

#   visualisation tools

class Plots:

    def __init__(self, MirNet, KeyNodesExtractor, miR_name):
        self.miR_G = MirNet.LCC
        self.key_nodes = KeyNodesExtractor.key_nodes
        self.miR_name = miR_name
        self.centrality_node = MirNet.get_LCCnd_centrality()
        self.card_LCC = KeyNodesExtractor.graph_features['card_LCC']
        self.n_CC = KeyNodesExtractor.graph_features['n_CC']
        self.idx_max_dy = KeyNodesExtractor.graph_features['cutoff_point']

    def central_distr(self):
        miR_G = self.miR_G
        key_nodes = self.key_nodes
        mir_name = self.miR_name

        """

        :param miR_G: a Graph of miRNA
        :param key_nodes: key_nodes from MainMiRNetwork.key_nodes
        :param mir_name: miR name for saving file
        :return: visualisation hist of centrality distribution
        """

        fig = plt.figure()
        ax = fig.add_subplot()
        N, bins, patches = ax.hist([miR_G.nodes[node]['BtweenCentrl'] for node in miR_G.nodes])
        ax.set(xlabel='Centrality',
               ylabel='Count of nodes')
        ax.yaxis.label.set_size(30)
        ax.xaxis.label.set_size(30)
        ax.tick_params(labelsize=20)

        ax.axvline(key_nodes[list(key_nodes.keys())[-1]], color='red', lw='4')

        right_side = ax.spines["right"]
        right_side.set_visible(False)
        top_side = ax.spines["top"]
        top_side.set_visible(False)

        # create gradient (grey_to_red hist path
        grey = Color('#cccccc')
        colors = list(grey.range_to(Color("red"), len(bins) - 1))
        for tmp_color, tmp_patch in zip(colors, patches):
            color = str(tmp_color)
            if len(color) < 7 and color[0] == '#':
                color = color + (7 - len(color)) * color[len(color) - 1]
            tmp_patch.set_facecolor(color)

        fig.set_figwidth(8.5)
        fig.set_figheight(8.5)

        plt.tight_layout()

        plt.savefig('./result/' + mir_name + '_centrality_distr.png', dpi=250)

    def graph_to_cytoscape(self):

        miR_G = nx.Graph(self.miR_G)  # unfreezing of the graph
        centrality_node = self.centrality_node
        rem_CC = 0

        for CC in list(nx.connected_components(miR_G)):
            if len(CC) < 3:
                miR_G.remove_nodes_from(CC)
                rem_CC += 1

        print(rem_CC, ' network components with less than two nodes have been removed', end='\n')

        PORT_NUMBER = 1234
        IP = 'localhost'
        BASE = 'http://' + IP + ':' + str(PORT_NUMBER) + '/v1/'

        requests.delete(BASE + 'session')  # Delete all networks in current session

        cytoscape_network = cy.from_networkx(miR_G)
        cytoscape_network['data']['name'] = 'miR_Net'
        res1 = requests.post(BASE + 'networks', data=json.dumps(cytoscape_network))
        res1_dict = res1.json()
        new_suid = res1_dict['networkSUID']
        requests.get(BASE + 'apply/layouts/force-directed/' + str(new_suid))

        # load and apply style

        res = requests.get(BASE + 'styles/miR_Net_Styles')
        if res.status_code != 200:

            with open('./options/cytoscape_styles/miR_Net_Styles.json') as json_file:
                miR_Net_Styles = json.load(json_file)

            for mapings in range(0, len(miR_Net_Styles['mappings'])):
                if miR_Net_Styles['mappings'][mapings]['visualProperty'] == 'NODE_LABEL_FONT_SIZE':
                    miR_Net_Styles['mappings'][mapings]['points'][1]['value'] = max(centrality_node.values())
                if miR_Net_Styles['mappings'][mapings]['visualProperty'] == 'NODE_SIZE':
                    miR_Net_Styles['mappings'][mapings]['points'][1]['value'] = max(centrality_node.values())
                if miR_Net_Styles['mappings'][mapings]['visualProperty'] == 'NODE_FILL_COLOR':
                    miR_Net_Styles['mappings'][mapings]['points'][1]['value'] = max(centrality_node.values())

            # Create new Visual Style
            res = requests.post(BASE + "styles", data=json.dumps(miR_Net_Styles))

        # Apply it to current network

        requests.get(
            BASE + 'apply/styles/' + 'miR_Net_Styles' + '/' + str(new_suid))  # !Это говно почему-то не работает

    def key_nodes_extractor(self):

        card_LCC = self.card_LCC
        n_CC = self.n_CC
        idx_max_dy = self.idx_max_dy
        mir_name = self.miR_name
        """

        :param card_LCC:
        :param n_CC:
        :param idx_max_dy:
        :param mir_name:
        :return: visualisation plot of key nodes selection
        """

        fig = plt.figure()
        ax = fig.add_subplot()
        ax.plot(card_LCC, linewidth=4, label='Cardinality of the LCC')
        ax.plot(n_CC, linewidth=4, label='Count of CC', color='tab:green', linestyle='dashed')
        # ax.plot(idx_max_dy, card_LCC[idx_max_dy], marker='o', markersize=20, color="red")
        ax.axvline(idx_max_dy, color='red', lw='4')
        ax.minorticks_on()
        ax.grid(which='major',
                color='w',
                linewidth=1.3)
        ax.grid(which='minor',
                color='w',
                linestyle=':')
        ax.set(xlabel='Number of top nodes removed',
               ylabel='LCC cardinality / Count of CC')
        ax.legend()

        right_side = ax.spines["right"]
        right_side.set_visible(False)
        top_side = ax.spines["top"]
        top_side.set_visible(False)

        #    plt.rc('font', size=10)  # controls default text sizes
        plt.rc('axes', labelsize=30)  # fontsize of the x and y labels
        plt.rc('xtick', labelsize=20)  # fontsize of the tick labels
        plt.rc('ytick', labelsize=20)  # fontsize of the tick labels
        plt.rc('legend', fontsize=20)  # legend fontsize
        fig.set_figwidth(8)
        fig.set_figheight(8)
        plt.tight_layout()

        plt.savefig('./result/' + mir_name + 'key_nodes_selection.png', dpi=300)



from reactome2py import analysis
import requests  # HTTP Client for Python
import json  # Standard JSON library
from py2cytoscape import util as cy
import networkx as nx
import pandas as pd
import io
import itertools as itt
from scipy.spatial.distance import pdist
import scipy.cluster
import matplotlib.pyplot as plt
from colour import Color
import numpy as np
from importlib.resources import files

def reactome_enrichment(gene_set, species='Homo sapiens'):
    target_set = ",".join(gene_set)

    result = analysis.identifiers(ids=target_set)
    token = result['summary']['token']

    token_result = analysis.token(token, species=species, page_size='-1', page='-1', sort_by='ENTITIES_FDR',
                                  order='ASC', resource='TOTAL', p_value='0.05', include_disease=False,
                                  min_entities=2, max_entities=None)

    return token_result


def reac_pars(token_result):

    class ReactomeRes:
        pass

    ReactomeRes.token = token_result['summary']['token']
    ReactomeRes.max_p_value = 0
    ReactomeRes.min_p_value = 99999
    ReactomeRes.max_ratio = 0
    ReactomeRes.min_ratio = 99999
    ReactomeRes.min_found = 99999
    ReactomeRes.max_found = 0
    ReactomeRes.react_dict = dict()

    for item in token_result['pathways']:

        if item['entities']['found'] < 2:
            continue

        ReactomeRes.react_dict[item['stId']] = {'name': item['name'],
                                      'fdr': item['entities']['fdr'],
                                      'p_value': item['entities']['pValue'],
                                      'found': item['entities']['found'],
                                      'ration': item['entities']['ratio']}

        #   for Networks

        if item['entities']['found'] < ReactomeRes.min_found:
            ReactomeRes.min_found = item['entities']['found']
        if item['entities']['found'] > ReactomeRes.max_found:
            ReactomeRes.max_found = item['entities']['found']

        if item['entities']['ratio'] < ReactomeRes.min_ratio:
            ReactomeRes.min_ratio = item['entities']['ratio']
        if item['entities']['ratio'] > ReactomeRes.max_ratio:
            ReactomeRes.max_ratio = item['entities']['ratio']

        if item['entities']['pValue'] < ReactomeRes.min_p_value:
            ReactomeRes.min_p_value = item['entities']['ratio']
        if item['entities']['pValue'] > ReactomeRes.max_p_value:
            ReactomeRes.max_p_value = item['entities']['ratio']

    return ReactomeRes


def get_net(ReactomeRes):

    reactome_linkege = pd.read_csv(files("genelens").joinpath("data/miRNET/baseData/ReactomePathwaysRelation.txt"), sep='\t', header=None)
    reactome_linkege.columns = ['Source', 'Target']

    G = nx.from_pandas_edgelist(reactome_linkege, source='Source', target='Target')
    G_enrich = G.subgraph(ReactomeRes.react_dict.keys())

    # removing network components that contain less than two nodses

    G_enrich = nx.Graph(G_enrich)

    for CC in list(nx.connected_components(G_enrich)):
        if len(CC) < 3:
            G_enrich.remove_nodes_from(CC)

    # add Statistics to Network

    for nods, dct in G_enrich.nodes(data=True):
        dct['label'] = ReactomeRes.react_dict[nods]['name']
        dct['fdr'] = ReactomeRes.react_dict[nods]['fdr']
        dct['found'] = ReactomeRes.react_dict[nods]['found']
        dct['ration'] = ReactomeRes.react_dict[nods]['ration']
        dct['p_value'] = ReactomeRes.react_dict[nods]['p_value']

    return G_enrich


def dendro_reactome_to_pandas(ReactomeRes, G, species='Homo sapiens'):

    url = 'https://reactome.org/AnalysisService/download/' + ReactomeRes.token + \
          '/pathways/TOTAL/result.csv'
    res = requests.get(url).content
    reactome_df = pd.read_csv(io.StringIO(res.decode('utf-8')))
    reactome_df = reactome_df[reactome_df['Entities pValue'] < 0.05]
    if (reactome_df['#Entities found'] >= 2).shape[0]:
        reactome_df = reactome_df[reactome_df['#Entities found'] >= 2]
    reactome_df = reactome_df[reactome_df['Species name'] == species]
    reactome_df = reactome_df[reactome_df['Pathway identifier'].isin(G.nodes)]

    return reactome_df, res

"""
def draw_net_to_cytoscape(G, ReactomeRes):

    PORT_NUMBER = 1234
    IP = 'localhost'
    BASE = 'http://' + IP + ':' + str(PORT_NUMBER) + '/v1/'

    # requests.delete(BASE + 'session')  # Delete all networks in current session

    cytoscape_network = cy.from_networkx(G)
    cytoscape_network['data']['name'] = 'First_Enrich_Net'
    res1 = requests.post(BASE + 'networks', data=json.dumps(cytoscape_network))
    res1_dict = res1.json()
    new_suid = res1_dict['networkSUID']
    requests.get(BASE + 'apply/layouts/force-directed/' + str(new_suid))

    # load and apply style

    res = requests.get(BASE + 'styles/PathwayEnrichStyles')
    if res.status_code != 200:

        with open('./options/cytoscape_styles/PathwayEnrichStyles.json') as json_file:
            directed_styles = json.load(json_file)

        for mapings in range(0, len(directed_styles['mappings'])):
            if directed_styles['mappings'][mapings]['visualProperty'] == 'NODE_FILL_COLOR':
                directed_styles['mappings'][mapings]['points'][0]['value'] = ReactomeRes.min_p_value
                directed_styles['mappings'][mapings]['points'][2]['value'] = 0.05
            if directed_styles['mappings'][mapings]['visualProperty'] == 'NODE_TRANSPARENCY':
                directed_styles['mappings'][mapings]['points'][0]['value'] = ReactomeRes.min_found
                directed_styles['mappings'][mapings]['points'][1]['value'] = ReactomeRes.max_found
            if directed_styles['mappings'][mapings]['visualProperty'] == 'NODE_LABEL_TRANSPARENCY':
                directed_styles['mappings'][mapings]['points'][0]['value'] = ReactomeRes.min_found
                directed_styles['mappings'][mapings]['points'][1]['value'] = ReactomeRes.max_found
            if directed_styles['mappings'][mapings]['visualProperty'] == 'NODE_BORDER_TRANSPARENCY':
                directed_styles['mappings'][mapings]['points'][0]['value'] = ReactomeRes.min_found
                directed_styles['mappings'][mapings]['points'][1]['value'] = ReactomeRes.max_found
            if directed_styles['mappings'][mapings]['visualProperty'] == 'NODE_SIZE':
                directed_styles['mappings'][mapings]['points'][0]['value'] = ReactomeRes.min_ratio
                directed_styles['mappings'][mapings]['points'][1]['value'] = ReactomeRes.max_ratio

        # Create new Visual Style
        res = requests.post(BASE + "styles", data=json.dumps(directed_styles))
        # new_style_name = res.json()['title']

        # Apply it to current network

    requests.get(
        BASE + 'apply/styles/' + 'PathwayEnrichStyles' + '/' + str(new_suid))  # !Это говно почему-то не работает

    return res
"""

# Dendrograms


def create_similarity_matrix(gene_sets):
        """Create a similarity matrix for a given pathway-geneset dataset.
        :param dict gene_sets: pathway gene set dictionary
        :rtype: pandas.DataFrame
        :returns: similarity matrix
        """
        index = sorted(gene_sets.keys())
        similarity_dataframe = pd.DataFrame(0.0, index=index, columns=index)

        for pathway_1, pathway_2 in itt.product(index, index):
            intersection = len(gene_sets[pathway_1].intersection(gene_sets[pathway_2]))
            smaller_set = min(len(gene_sets[pathway_1]), len(gene_sets[pathway_2]))

            similarity = float(intersection / smaller_set)  # Formula to calculate similarity

            #similarity_dataframe[pathway_1][pathway_2] = similarity
            similarity_dataframe.loc[pathway_2, pathway_1] = similarity

        return similarity_dataframe


def get_dendro(dt, key_nodes_dict, fig_preff_name=''):

    gene_set_dict = {dt.iloc[row]['Pathway identifier']: set(dt.iloc[row]['Submitted entities found'].split(';')) for
                     row in range(0, dt.shape[0])}
    dtid2name = {dt.iloc[row]['Pathway identifier']: set(dt.iloc[row]['Pathway name'].split(';')) for row in
                 range(0, dt.shape[0])}

    similarity_matrix = create_similarity_matrix(gene_set_dict)
    if similarity_matrix.empty:
        print('similarity_matrix is empty. No intersections found. Dendrogram could not be constructed.')
        return dt
    distance_matrix = pdist(similarity_matrix, metric='correlation')
    # Замена NaN на 0
    distance_matrix = np.nan_to_num(distance_matrix, nan=0.0)

    # Cluster hierarchicaly using scipy
    clusters = scipy.cluster.hierarchy.linkage(distance_matrix, method='single')
    T = scipy.cluster.hierarchy.to_tree(clusters, rd=False)

    # Create dictionary for labeling nodes by their IDs
    labels = list(similarity_matrix.columns)
    id2name = dict(zip(range(len(labels)), labels))

    def getlabel(x):
        return '{}: ({})'.format(''.join(dtid2name[id2name[x]]), ', '.join(gene_set_dict[id2name[x]]))

    # Draw dendrogram using matplotlib to scipy-dendrogram.pdf
    my_dpi = 10
    plt.figure(figsize=(2400 / my_dpi, 2400 / my_dpi), dpi=my_dpi)
    fig, axes = plt.subplots()
    scipy.cluster.hierarchy.dendrogram(clusters,
                                       leaf_label_func=getlabel,
                                       orientation='right')
    fig.set_figwidth(10)
    fig.set_figheight(16)
    plt.tight_layout()

    """
    set a label color as the normalised cumulative betweenness {sum(betweenness)/len(gene_set)}
    """
    top5centralNods = key_nodes_dict

    ylbls = axes.get_ymajorticklabels()

    # create label to normalised cumulative betweenness dictionary

    label_to_BC = dict()
    for lbl in ylbls:
        local_label = (lbl.get_text().split(': ('))[0]  # get path name
        gene_set = (lbl.get_text().split(': ('))[1][:-1].split(', ')  # get gene set
        for gene in gene_set:
            if gene not in top5centralNods:
                top5centralNods[gene] = 0
        sum_BC = sum([top5centralNods[gene] for gene in gene_set]) / len(gene_set)
        label_to_BC[local_label] = sum_BC

    label_to_BC_sort = {k: v for k, v in sorted(label_to_BC.items(), key=lambda item: item[1], reverse=False)}

    # create gradient (grey_to_red hist path
    grey = Color('grey')
    colors = list(grey.range_to(Color("red"), len(label_to_BC_sort)))
    for label, tmp_color in zip(label_to_BC_sort.keys(), colors):
        color = str(tmp_color)
        if len(color) < 7 and color[0] == '#':
            color = color + (7 - len(color)) * color[len(color) - 1]
        label_to_BC_sort[label] = color

    # apply gradient to labels
    for lbl in ylbls:
        local_label = (lbl.get_text().split(': ('))[0]
        lbl.set_color(label_to_BC_sort[local_label])

    plt.savefig("./results/" + fig_preff_name + "scipy_dendrogram.png", dpi=300)
